/*
 * @(#)BasicScrollBarUI.java	1.49 98/08/26
 *
 * Copyright 1997, 1998 by Sun Microsystems, Inc.,
 * 901 San Antonio Road, Palo Alto, California, 94303, U.S.A.
 * All rights reserved.
 *
 * This software is the confidential and proprietary information
 * of Sun Microsystems, Inc. ("Confidential Information").  You
 * shall not disclose such Confidential Information and shall use
 * it only in accordance with the terms of the license agreement
 * you entered into with Sun.
 *
 */
  
package fi.helsinki.cs.gist.scroll;

import fi.helsinki.cs.gist.*;
import java.util.Vector;
import java.util.ListIterator;
import javax.swing.text.*;
import javax.swing.*;

import java.awt.*;
import java.awt.event.*;

import java.beans.*;

import javax.swing.*;
import javax.swing.event.*;
import javax.swing.plaf.*;
import javax.swing.plaf.basic.*;


/**
 * Basic implementation of the PlaceholderUI
 *
 * PlaceholderUI written by
 * @author asaura
 * @version 17.8.1999
 *
 * based on BasicScrollBarUI
 *
 * @version 1.49 08/26/98
 * @author Rich Schiavi
 * @author David Kloba
 * @author Hans Muller
 */

public class PlaceholderUI 
    extends ScrollBarUI implements LayoutManager, SwingConstants,
						  TagAreaListener
{
    protected Dimension minimumThumbSize;
    protected Dimension maximumThumbSize;
  
    protected Color thumbHighlightColor;
    protected Color thumbLightShadowColor;
    protected Color thumbDarkShadowColor;
    protected Color thumbColor;
    protected Color trackColor;
    protected Color trackHighlightColor;
    // GIST / Karri
    protected Color shadeColor;
    protected static final int SHADE_SQUARE_SIZE = 3; 
  
    protected Placeholder scrollbar;
    protected JButton incrButton;
    protected JButton decrButton;
    // GIST / AS & Karri
    protected JButton lesserIncrButton;
    protected JButton lesserDecrButton;  
    protected JButton newTagButton;
    protected JToggleButton showHideButton;
    protected TagArea tagArea;

    protected boolean isDragging;
    protected TrackListener trackListener;
    protected ArrowButtonListener buttonListener;
    protected NewTagButtonListener newTagButtonListener;
    protected ShowHideButtonListener showHideButtonListener;
    protected ModelListener modelListener;
  
    protected Rectangle thumbRect;
    protected Rectangle trackRect;
  
    protected int trackHighlight;
  
    protected static final int NO_HIGHLIGHT = 0;
    protected static final int DECREASE_HIGHLIGHT = 1;
    protected static final int INCREASE_HIGHLIGHT = 2;
  
    protected ScrollListener scrollListener;
    protected PropertyChangeListener propertyChangeListener; 
    protected Timer scrollTimer;
  
  
    public static ComponentUI createUI(JComponent c)    {
	return new PlaceholderUI();
    }
  
  
    protected void configureScrollBarColors() 
	{
	    thumbHighlightColor = UIManager.getColor("ScrollBar.thumbHighlight");
	    thumbLightShadowColor = UIManager.getColor("ScrollBar.thumbLightShadow");
	    thumbDarkShadowColor = UIManager.getColor("ScrollBar.thumbDarkShadow");
	    thumbColor = UIManager.getColor("ScrollBar.thumb");
	    trackColor = UIManager.getColor("ScrollBar.track");
	    trackHighlightColor = UIManager.getColor("ScrollBar.trackHighlight");
	    // GIST / Karri
	    shadeColor = new Color(140, 140, 140);
	}


    public void installUI(JComponent c)   {
	scrollbar = (Placeholder)c;
        thumbRect = new Rectangle(0, 0, 0, 0);
        trackRect = new Rectangle(0, 0, 0, 0);
	installDefaults();
	installComponents();
	installListeners();
	installKeyboardActions();
    }

    public void uninstallUI(JComponent c) {
	scrollbar = (Placeholder)c;
	uninstallDefaults();
	uninstallComponents();
	uninstallListeners();
	uninstallKeyboardActions();
	c.remove(incrButton);
	c.remove(decrButton);
	c.remove(lesserIncrButton);
	c.remove(lesserDecrButton);
	c.setLayout(null);
	thumbRect = null;
	scrollbar = null;
	incrButton = null;
	decrButton = null;
	lesserIncrButton = null;
	lesserDecrButton = null;
    }
    

    protected void installDefaults() {
	minimumThumbSize = (Dimension)UIManager.get("ScrollBar.minimumThumbSize");
	maximumThumbSize = (Dimension)UIManager.get("ScrollBar.maximumThumbSize");

	trackHighlight = NO_HIGHLIGHT;
	scrollbar.setLayout(this);

	switch (scrollbar.getOrientation()) {
	case Placeholder.VERTICAL:
	    incrButton = createIncreaseButton(SOUTH);
	    decrButton = createDecreaseButton(NORTH);
	    lesserIncrButton  = createIncreaseButton(SOUTH);
	    lesserDecrButton  = createDecreaseButton(NORTH);
	    tagArea = new TagArea();
	    tagArea.addTagAreaListener(this);
	    Icon newTagIcon = new ImageIcon(
		"fi/helsinki/cs/gist/images/EnabledNewTagIcon.gif");
	    newTagButton = new JButton(newTagIcon);
            newTagButton.setToolTipText("Create new tag");
	    
	    ImageIcon showHideToggleIcon = new ImageIcon(
		"fi/helsinki/cs/gist/images/ShowHideToggleIcon.gif");
	    showHideButton = new JToggleButton(showHideToggleIcon, true);
            showHideButton.setToolTipText("Hide tags");

	    Insets buttonInsets = new Insets(1,0,1,0);
	    newTagButton.setMargin(buttonInsets);
	    showHideButton.setMargin(buttonInsets);
	    
	    scrollbar.add(incrButton);
	    scrollbar.add(decrButton);
	    scrollbar.add(lesserIncrButton);
	    scrollbar.add(lesserDecrButton);
	    scrollbar.add(newTagButton);
	    scrollbar.add(tagArea);
	    scrollbar.add(showHideButton);
	    break;
            
	case Placeholder.HORIZONTAL:
	    incrButton = createIncreaseButton(EAST);
	    decrButton = createDecreaseButton(WEST);
	    scrollbar.add(incrButton);
	    scrollbar.add(decrButton);
	    break;
	}
                
	scrollbar.setEnabled(scrollbar.isEnabled());
	scrollbar.setOpaque(true);
	configureScrollBarColors();
	LookAndFeel.installBorder(scrollbar, "ScrollBar.border");
    }


    protected void installComponents(){
    }

    protected void uninstallComponents(){
    }

  
    protected void installListeners(){
	trackListener = createTrackListener();
    	buttonListener = createArrowButtonListener();
    	modelListener = createModelListener();
	propertyChangeListener = createPropertyChangeListener();
	newTagButtonListener = createNewTagButtonListener();
	showHideButtonListener = createShowHideButtonListener();
	
	scrollbar.addMouseListener(trackListener);
	scrollbar.addMouseMotionListener(trackListener);
        scrollbar.getModel().addChangeListener(modelListener);
	scrollbar.addPropertyChangeListener(propertyChangeListener);

        if (incrButton != null) {
            incrButton.addMouseListener(buttonListener);
	}
        if (decrButton != null)	{
            decrButton.addMouseListener(buttonListener);
	}
        if (lesserIncrButton != null) {
            lesserIncrButton.addMouseListener(buttonListener);
	}
        if (lesserDecrButton != null)	{
            lesserDecrButton.addMouseListener(buttonListener);
	}
	if (newTagButton != null) {
	    newTagButton.addActionListener(newTagButtonListener);
	}
	if (showHideButton != null) {
	    showHideButton.addActionListener(showHideButtonListener);
	}

	scrollListener = createScrollListener();
	scrollTimer = new Timer(20, scrollListener);
	scrollTimer.setInitialDelay(300);  // default InitialDelay?
    }


    protected void installKeyboardActions(){
    }

    protected void uninstallKeyboardActions(){
    }


    protected void uninstallListeners() {
	scrollTimer.stop();
	scrollTimer = null;

	if (decrButton != null){
	    decrButton.removeMouseListener(buttonListener);
	}
	if (incrButton != null){
	    incrButton.removeMouseListener(buttonListener);
	}
	if (lesserDecrButton != null){
	    lesserDecrButton.removeMouseListener(buttonListener);
	}
	if (lesserIncrButton != null){
	    lesserIncrButton.removeMouseListener(buttonListener);
	}
	if (newTagButton != null){
	    newTagButton.removeActionListener(newTagButtonListener);
	}
	if (showHideButton != null){
	    showHideButton.removeActionListener(showHideButtonListener);
	}

	scrollbar.getModel().removeChangeListener(modelListener);
	scrollbar.removeMouseListener(trackListener);
	scrollbar.removeMouseMotionListener(trackListener);
	scrollbar.removePropertyChangeListener(propertyChangeListener);
    }


    protected void uninstallDefaults(){
        LookAndFeel.uninstallBorder(scrollbar);
    }


    protected TrackListener createTrackListener(){
	return new TrackListener();
    }

    protected ArrowButtonListener createArrowButtonListener(){
	return new ArrowButtonListener();
    }
    
    protected ModelListener createModelListener(){
	return new ModelListener();
    }

    protected ScrollListener createScrollListener(){
	return new ScrollListener();
    }
    
    protected PropertyChangeListener createPropertyChangeListener() {
	return new PropertyChangeHandler();
    }

    protected NewTagButtonListener createNewTagButtonListener() {
	return new NewTagButtonListener();
    }
  
    protected ShowHideButtonListener createShowHideButtonListener() {
	return new ShowHideButtonListener();
    }
  
    public void paint(Graphics g, JComponent c) {
	paintTrack(g, c, getTrackBounds());		
	paintThumb(g, c, getThumbBounds());
    }

        
    /**
      * A vertical scrollbars preferred width is the maximum of 
      * preferred widths of the (non null) increment/decrement buttons,
      * and the minimum width of the thumb. The preferred height is the 
      * sum of the preferred heights of the same parts.  The basis for 
      * the preferred size of a horizontal scrollbar is similar. 
      * <p>
      * The preferredSize is only computed once, subequent
      * calls to this method just return a cached size.  T
      * 
      * @param c The JScrollBar that's delegating this method to us.
      * @return The preferred size of a Basic JScrollBar.
      * @see #getMaximumSize
      * @see #getMinimumSize
      */
    public Dimension getPreferredSize(JComponent c) {
	return (scrollbar.getOrientation() == Placeholder.VERTICAL)
	    ? new Dimension(16, 48)
	    : new Dimension(48, 16);
    }


    /**
      * A vertical scrollbars minimum width is the largest 
      * minimum width of the (non null) increment/decrement buttons,
      * and the minimum width of the thumb. The minimum height is the 
      * sum of the minimum heights of the same parts.  The basis for 
      * the preferred size of a horizontal scrollbar is similar. 
      * <p>
      * The minimumSize is only computed once, subequent
      * calls to this method just return a cached size.  T
      * 
      * @param c The JScrollBar that's delegating this method to us.
      * @return The minimum size of a Basic JScrollBar.
      * @see #getMaximumSize
      * @see #getPreferredSize
      */
    public Dimension getMinimumSize(JComponent c) {
	return getPreferredSize(c);
    }

    
    /**
      * @param c The JScrollBar that's delegating this method to us.
      * @return new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE);
      * @see #getMinimumSize
      * @see #getPreferredSize
      */
    public Dimension getMaximumSize(JComponent c) {
        return new Dimension(Integer.MAX_VALUE, Integer.MAX_VALUE);
    }
    	
    
    protected JButton createDecreaseButton(int orientation)  {
        return new BasicArrowButton(orientation);
    }

    protected JButton createIncreaseButton(int orientation)  {
        return new BasicArrowButton(orientation);
    }
          		  

    protected void paintDecreaseHighlight(Graphics g)
	{
	    Insets insets = scrollbar.getInsets();
	    Rectangle thumbR = getThumbBounds();
	    g.setColor(trackHighlightColor);

	    if (scrollbar.getOrientation() == Placeholder.VERTICAL) {
		int x = insets.left;
		int y = lesserIncrButton.getY() + lesserIncrButton.getHeight();
		int w = scrollbar.getWidth() - (insets.left + insets.right);
		int h = thumbR.y - y;
		g.fillRect(x, y, w, h);
	    } 
	    else	{
		int x = decrButton.getX() + decrButton.getHeight();
		int y = insets.top;
		int w = thumbR.x - x;
		int h = scrollbar.getHeight() - (insets.top + insets.bottom);
		g.fillRect(x, y, w, h);
	    }
	}      
	

    protected void paintIncreaseHighlight(Graphics g)
	{
	    Insets insets = scrollbar.getInsets();
	    Rectangle thumbR = getThumbBounds();
	    g.setColor(trackHighlightColor);

	    if (scrollbar.getOrientation() == Placeholder.VERTICAL) {
		int x = insets.left;
		int y = thumbR.y + thumbR.height;
		int w = scrollbar.getWidth() - (insets.left + insets.right);
		int h = lesserDecrButton.getY() - y;
		g.fillRect(x, y, w, h);
	    } 
	    else {
		int x = thumbR.x + thumbR.width;
		int y = insets.top;
		int w = incrButton.getX() - x;
		int h = scrollbar.getHeight() - (insets.top + insets.bottom);
		g.fillRect(x, y, w, h);
	    }
	}      


    protected void paintTrack(Graphics g, JComponent c, Rectangle trackBounds)  
	{
	    g.setColor(trackColor);
	    // GIST / Karri
	    // First the background
	    g.fillRect(trackBounds.x, trackBounds.y, trackBounds.width, trackBounds.height);

	    // Then the square shading
	    ListIterator i = scrollbar.getShadedAreas().listIterator();
	    while (i.hasNext()) {
		Rectangle r = (Rectangle) i.next();
		int y = getVisibleY(r.y);
		int height = getVisibleY(r.y + r.height) - y;
		    
		if (y + height < trackBounds.height) { // Otherwise insane
		    boolean startWithPaint = true; // false;
		    // int m = r.x % (2 * SHADE_SQUARE_SIZE);
		    // if (m < SHADE_SQUARE_SIZE)
		    //    startWithPaint = true;
		    // m %= SHADE_SQUARE_SIZE;
		    
		    g.setColor(shadeColor);
		    
		    while (height > 0) {
			boolean paint = startWithPaint;
			int x = trackBounds.x;
			int width = trackBounds.width;
			int h = height > SHADE_SQUARE_SIZE ? SHADE_SQUARE_SIZE : height;
			// Give an impression of a continuous
			// squared region by adjusting the height
			// of the first square row.
			// if (m != 0 && h > m)
			//     h -= m;

			while (width > 0) {
			    int w = width > SHADE_SQUARE_SIZE ? SHADE_SQUARE_SIZE : width;
			    if (paint) 
				g.fillRect(x, y, w, h);
			    paint = !paint;
			    x += SHADE_SQUARE_SIZE;
			    width -= w;
			}
			y += SHADE_SQUARE_SIZE; // - m;
			// m = 0;
			height -= h;
			startWithPaint = !startWithPaint;
		    }
		}
	    }

	    if(trackHighlight == DECREASE_HIGHLIGHT)	{
		paintDecreaseHighlight(g);
	    } 
	    else if(trackHighlight == INCREASE_HIGHLIGHT)		{
		paintIncreaseHighlight(g);
	    }
	}

	
    protected void paintThumb(Graphics g, JComponent c, Rectangle thumbBounds)  
	{
	    if(thumbBounds.isEmpty() || !scrollbar.isEnabled())	{
		return;
	    }

	    int w = thumbBounds.width;
	    int h = thumbBounds.height;		

	    g.translate(thumbBounds.x, thumbBounds.y);

	    g.setColor(thumbDarkShadowColor);
	    g.drawRect(0, 0, w-1, h-1);    
	    g.setColor(thumbColor);
	    g.fillRect(0, 0, w-1, h-1);
        
	    g.setColor(thumbHighlightColor);
	    g.drawLine(1, 1, 1, h-2);
	    g.drawLine(2, 1, w-3, 1);
        
	    g.setColor(thumbLightShadowColor);
	    g.drawLine(2, h-2, w-2, h-2);
	    g.drawLine(w-2, 1, w-2, h-3);

	    g.translate(-thumbBounds.x, -thumbBounds.y);
	}


    /** 
      * Return the smallest acceptable size for the thumb.  If the scrollbar
      * becomes so small that this size isn't available, the thumb will be
      * hidden.  
      * <p>
      * <b>Warning </b>: the value returned by this method should not be
      * be modified, it's a shared static constant.
      *
      * @return The smallest acceptable size for the thumb.
      * @see #getMaximumThumbSize
      */
    protected Dimension getMinimumThumbSize() { 
	return minimumThumbSize;
    }

    /** 
      * Return the smallest acceptable size for the thumb.  If the scrollbar
      * becomes so small that this size isn't available, the thumb will be
      * hidden.  To create a fixed size thumb one make this
      * method and <code>getMinimumThumbSize</code> return the same value.
      * <p>
      * <b>Warning </b>: the value returned by this method should not be
      * be modified, it's a shared static constant.
      *
      * @return The smallest acceptable size for the thumb.
      * @see #getMinimumThumbSize
      */
    protected Dimension getMaximumThumbSize()	{ 
	return maximumThumbSize;
    }


    /*
     * LayoutManager Implementation
     */

    public void addLayoutComponent(String name, Component child) {}
    public void removeLayoutComponent(Component child) {}
    
    public Dimension preferredLayoutSize(Container scrollbarContainer) {
        return getPreferredSize((JComponent)scrollbarContainer);
    }
    
    public Dimension minimumLayoutSize(Container scrollbarContainer) {
        return getMinimumSize((JComponent)scrollbarContainer);
    }
    

    protected void layoutVScrollbar(Placeholder sb) {
	Dimension sbSize = sb.getSize();
	Insets sbInsets = sb.getInsets();

	int tagAreaW = 16;
	
	/*
	 * Width and left edge of the buttons and thumb.
	 */
	int itemW = sbSize.width - tagAreaW - (sbInsets.left + sbInsets.right);
	int itemX = sbInsets.left;
        
	/* Nominal locations of the buttons, assuming their preferred
	 * size will fit.
	 */
	int decrButtonH = decrButton.getPreferredSize().height;
	int decrButtonY = sbInsets.top;
        
	int lesserIncrButtonH = incrButton.getPreferredSize().height;
	int lesserIncrButtonY = decrButtonY + decrButtonH;

	int incrButtonH = incrButton.getPreferredSize().height;
	int incrButtonY = sbSize.height - (sbInsets.bottom + incrButtonH);

	int lesserDecrButtonH = lesserDecrButton.getPreferredSize().height;
	int lesserDecrButtonY = incrButtonY - lesserDecrButtonH;
	
	int newTagButtonH = decrButtonH + lesserIncrButtonH;
	int newTagButtonY = sbInsets.top;
		
	int showHideButtonH = incrButtonH + lesserDecrButtonH;
	int showHideButtonY = sbSize.height -
	    (sbInsets.bottom +
	     incrButtonH + lesserDecrButtonH);
		
	/* The thumb must fit within the height left over after we
	 * subtract the preferredSize of the buttons and the insets.
	 */
	int sbInsetsH = sbInsets.top + sbInsets.bottom;
	int sbButtonsH = decrButtonH + lesserIncrButtonH +
	    lesserDecrButtonH + incrButtonH;
	float trackH = sbSize.height - (sbInsetsH + sbButtonsH);
        
	/* Compute the height and origin of the thumb.   The case
	 * where the thumb is at the bottom edge is handled specially 
	 * to avoid numerical problems in computing thumbY.  Enforce
	 * the thumbs min/max dimensions.  If the thumb doesn't
	 * fit in the track (trackH) we'll hide it later.
	 */
	float min = sb.getMinimum();
	float extent = sb.getVisibleAmount();
	float range = sb.getMaximum() - min;
	float value = sb.getValue();

	int thumbH = (range <= 0) 
	    ? getMaximumThumbSize().height : (int)(trackH * (extent / range));
	thumbH = Math.max(thumbH, getMinimumThumbSize().height);
	thumbH = Math.min(thumbH, getMaximumThumbSize().height);
        
	int thumbY = lesserDecrButtonY - thumbH;  
	if (sb.getValue() < (sb.getMaximum() - sb.getVisibleAmount())) {
	    float thumbRange = trackH - thumbH;
	    thumbY = (int)(0.5f + (thumbRange * ((value - min) / (range - extent))));
	    thumbY +=  lesserIncrButtonY + lesserIncrButtonH;
	}

	// tags
	for(int i = 0; i < tagArea.tags.size(); i++) {
	    Tag t = (Tag)tagArea.tags.elementAt(i);
	    float thumbRange = trackH - thumbH;
	    int tagY = (int)(0.5f + (thumbRange * ((t.value - min) / (range - extent))));
	    t.setDrawnY(tagY + thumbH / 2);
	}
	
	/* If the buttons don't fit, allocate a quarter of the available 
	 * space to each and move the lower ones (incrButton+ldb) down.
	 */
	int sbAvailButtonH = (sbSize.height - sbInsetsH);
	if (sbAvailButtonH < sbButtonsH) {
	    incrButtonH = decrButtonH =
		lesserIncrButtonH = lesserDecrButtonH = sbAvailButtonH / 4;
	    lesserDecrButtonY = sbSize.height -
		(sbInsets.bottom + lesserDecrButtonH);
	    incrButtonY = sbSize.height - (sbInsets.bottom + incrButtonH);
	}

	decrButton.setBounds(itemX, decrButtonY, itemW, decrButtonH);
	lesserIncrButton.setBounds(itemX, lesserIncrButtonY,
				   itemW, lesserIncrButtonH);	
	lesserDecrButton.setBounds(itemX, lesserDecrButtonY,
				   itemW, lesserDecrButtonH);
	incrButton.setBounds(itemX, incrButtonY, itemW, incrButtonH);

	/* Update the trackRect field.
	 */	
	int itrackY = lesserIncrButtonY + lesserIncrButtonH;
	int itrackH = lesserDecrButtonY - itrackY;
	trackRect.setBounds(itemX, itrackY, itemW, itrackH);
	
	tagArea.setBounds(itemX+itemW, newTagButtonY+newTagButtonH,
			  tagAreaW, itrackH);

	newTagButton.setBounds(itemX+itemW, newTagButtonY,
			       tagAreaW, newTagButtonH);
	showHideButton.setBounds(itemX+itemW, showHideButtonY,
				 tagAreaW, showHideButtonH);
	
	
	/* If the thumb isn't going to fit, zero it's bounds.  Otherwise
	 * make sure it fits between the buttons.  Note that setting the
	 * thumbs bounds will cause a repaint.
	 */
	if(thumbH >= (int)trackH)	{
	    setThumbBounds(0, 0, 0, 0);
	}
	else {
	    if ((thumbY + thumbH) > lesserDecrButtonY) {
		thumbY = lesserDecrButtonY - thumbH;
	    }
	    if (thumbY  < (lesserIncrButtonY + lesserIncrButtonH)) {
		thumbY = lesserIncrButtonY + lesserIncrButtonH + 1;
	    }
	    setThumbBounds(itemX, thumbY, itemW, thumbH);
	}
    }
    

    protected void layoutHScrollbar(Placeholder sb) {
	Dimension sbSize = sb.getSize();
	Insets sbInsets = sb.getInsets();
        
	/* Height and top edge of the buttons and thumb.
	 */
	int itemH = sbSize.height - (sbInsets.top + sbInsets.bottom);
	int itemY = sbInsets.top;

	/* Nominal locations of the buttons, assuming their preferred
	 * size will fit.
	 */
	int decrButtonW = decrButton.getPreferredSize().width;
	int decrButtonX = sbInsets.left;
        
	int incrButtonW = incrButton.getPreferredSize().width;
	int incrButtonX = sbSize.width - (sbInsets.right + incrButtonW);
        
	/* The thumb must fit within the width left over after we
	 * subtract the preferredSize of the buttons and the insets.
	 */
	int sbInsetsW = sbInsets.left + sbInsets.right;
	int sbButtonsW = decrButtonW + incrButtonW;
	float trackW = sbSize.width - (sbInsetsW + sbButtonsW);
        
	/* Compute the width and origin of the thumb.  Enforce
	 * the thumbs min/max dimensions.  The case where the thumb 
	 * is at the right edge is handled specially to avoid numerical 
	 * problems in computing thumbX.  If the thumb doesn't
	 * fit in the track (trackH) we'll hide it later.
	 */
	float min = sb.getMinimum();
	float extent = sb.getVisibleAmount();
	float range = sb.getMaximum() - min;
	float value = sb.getValue();

	int thumbW = (range <= 0) 
	    ? getMaximumThumbSize().width : (int)(trackW * (extent / range));
	thumbW = Math.max(thumbW, getMinimumThumbSize().width);
	thumbW = Math.min(thumbW, getMaximumThumbSize().width);
        
	int thumbX = incrButtonX - thumbW;
	if (sb.getValue() < (sb.getMaximum() - sb.getVisibleAmount())) {
	    float thumbRange = trackW - thumbW;
	    thumbX = (int)(0.5f + (thumbRange * ((value - min) / (range - extent))));
	    thumbX +=  decrButtonX + decrButtonW;
	}

	/* If the buttons don't fit, allocate half of the available 
	 * space to each and move the right one (incrButton) over.
	 */
	int sbAvailButtonW = (sbSize.width - sbInsetsW);
	if (sbAvailButtonW < sbButtonsW) {
	    incrButtonW = decrButtonW = sbAvailButtonW / 2;
	    incrButtonX = sbSize.width - (sbInsets.right + incrButtonW);
	}
        
	decrButton.setBounds(decrButtonX, itemY, decrButtonW, itemH);
	incrButton.setBounds(incrButtonX, itemY, incrButtonW, itemH);

	/* Update the trackRect field.
	 */	
	int itrackX = decrButtonX + decrButtonW;
	int itrackW = incrButtonX - itrackX;
	trackRect.setBounds(itrackX, itemY, itrackW, itemH);

	/* Make sure the thumb fits between the buttons.  Note 
	 * that setting the thumbs bounds causes a repaint.
	 */
	if (thumbW >= (int)trackW) {
	    setThumbBounds(0, 0, 0, 0);
	}
	else {
	    if (thumbX + thumbW > incrButtonX) {
		thumbX = incrButtonX - thumbW;
	    }
	    if (thumbX  < decrButtonX + decrButtonW) {
		thumbX = decrButtonX + decrButtonW + 1;
	    }
	    setThumbBounds(thumbX, itemY, thumbW, itemH);
	}
    }
    

    public void layoutContainer(Container scrollbarContainer) {
	/* If the user is dragging the value, we'll assume that the 
	 * scrollbars layout is OK modulo the thumb which is being 
	 * handled by the dragging code.
	     */
	if (isDragging) {
	    return;
	}

	Placeholder scrollbar = (Placeholder)scrollbarContainer;
	switch (scrollbar.getOrientation()) {
	case Placeholder.VERTICAL:
	    layoutVScrollbar(scrollbar);
	    break;
            
	case Placeholder.HORIZONTAL:
	    layoutHScrollbar(scrollbar);
	    break;
	}
    }


    /**
      * Set the bounds of the thumb and force a repaint that includes
      * the old thumbBounds and the new one.
      *
      * @see #getThumbBounds
      */
    protected void setThumbBounds(int x, int y, int width, int height) {
	/* If the thumbs bounds haven't changed, we're done.
	 */
	if ((thumbRect.x == x) && 
	    (thumbRect.y == y) && 
	    (thumbRect.width == width) && 
	    (thumbRect.height == height)) {
	    return;
	}

	/* Update thumbRect, and repaint the union of x,y,w,h and 
	 * the old thumbRect.
	 */
	int minX = Math.min(x, thumbRect.x);
	int minY = Math.min(y, thumbRect.y);
	int maxX = Math.max(x + width, thumbRect.x + thumbRect.width);
	int maxY = Math.max(y + height, thumbRect.y + thumbRect.height);

	thumbRect.setBounds(x, y, width, height);
	scrollbar.repaint(minX, minY, maxX - minX, maxY - minY);

	checkThumbOverlap();
    }

    /**
      * Return the current size/location of the thumb.
      * <p>
      * <b>Warning </b>: the value returned by this method should not be
      * be modified, it's a reference to the actual rectangle, not a copy.
      *
      * @return The current size/location of the thumb.
      * @see #setThumbBounds
      */
    protected Rectangle getThumbBounds() {
	return thumbRect;
    }


    /**
      * Return the current bounds of the track, i.e. the space in between
      * the increment and decrement buttons, less the insets.  The value
      * returned by this method is updated each time the scrollbar is
      * layed out (validated).
      * <p>
      * <b>Warning </b>: the value returned by this method should not be
      * be modified, it's a reference to the actual rectangle, not a copy.
      *
      * @return The current bounds of the scrollbar track.
      * @see #layoutContainer
      */
    protected Rectangle getTrackBounds() {
	return trackRect;
    }
  
    protected void scrollByBlock(int direction) {
	synchronized(scrollbar)	{
	    int oldValue = scrollbar.getValue();
	    int blockIncrement = scrollbar.getBlockIncrement(direction);
	    System.out.println("BlockIncrement: " + blockIncrement);
	    int delta = blockIncrement * ((direction > 0) ? +1 : -1);
      
	    scrollbar.setValue(oldValue + delta);			
	    trackHighlight = direction > 0 ? INCREASE_HIGHLIGHT : DECREASE_HIGHLIGHT;
	    Rectangle dirtyRect = getTrackBounds();
	    scrollbar.repaint(dirtyRect.x, dirtyRect.y, dirtyRect.width, dirtyRect.height);
	}
    }
    

    protected void scrollByUnit(int direction)	{
	synchronized(scrollbar)	{
	    int delta;
	    if(direction > 0)
		delta = scrollbar.getUnitIncrement(direction);
	    else
		delta = -scrollbar.getUnitIncrement(direction);
	    scrollbar.setValue(delta + scrollbar.getValue()); 
	}
    }

    // GIST / Karri
    public void setTags(Vector v) {
	tagArea.setTags(v);
    }

    public Vector getTags() {
	return tagArea.getTags();
    }

    /** Scroll to the location given, whichever direction it may be. */
    protected void scrollTo(int target) {
	System.err.println("PUI: scrolling to "+target);
	int currentValue = getThumbBounds().y;

	boolean goingUp = target < currentValue;

	BoundedRangeModel model = scrollbar.getModel();
	Rectangle thumbR = getThumbBounds();
	float trackLength;
	int thumbMin = 0;
	int thumbMax = 0;
	int thumbPos = 0;

	int offset = 0; // ?
    
	if (scrollbar.getOrientation() == Placeholder.VERTICAL) {
	    thumbMin = lesserIncrButton.getY() +
		lesserIncrButton.getHeight();
	    thumbMax = lesserDecrButton.getY() -
		getThumbBounds().height;
	    thumbPos = Math.min(thumbMax,
				Math.max(thumbMin, (target - offset)));
	    setThumbBounds(thumbR.x, thumbPos, thumbR.width, thumbR.height);
	    trackLength = getTrackBounds().height;
	}
	else {
	    /* nothing as of yet. This doesn't matter anyway, since
	       this *is* a vertical scrollbar.
	    */
	}

	/* Set the scrollbars value.  If the thumb has reached the end of
	 * the scrollbar, then just set the value to its maximum.  Otherwise
	 * compute the value as accurately as possible.
	 */
	if (thumbPos == thumbMax) {
	    scrollbar.setValue(model.getMaximum() - model.getExtent());
	}
	else {
	    float valueMax = model.getMaximum() - model.getExtent();
	    float valueRange = valueMax - model.getMinimum();
	    float thumbValue = thumbPos - thumbMin;
	    float thumbRange = thumbMax - thumbMin;
	    int value = (int)(0.5 + ((thumbValue / thumbRange) * valueRange));
	    scrollbar.setValue(value + model.getMinimum());
	}
    }

    //////////////////////////////////////////////////////////////////////
    // TagAreaListener implementation

    public void tagAdded(TagEvent e) {
	System.err.println("PUI: tag added event received");
	newTagButton.setEnabled(false);
    }

    public void tagDeleted(TagEvent e) {
	System.err.println("PUI: tag deleted event received");
	checkThumbOverlap();
    }

    public void tagChanged(TagEvent e) {
	System.err.println("PUI: tag changed event received");
    }

    public void tagActivated(TagEvent e) {
	System.err.println("PUI: tag request event received");
	scrollbar.setValue(e.getValue());
	
	// Recalculate the layout to set the thumb in the right place.
	// This is needed at least in drag snap, where the thumb
	// position cannot be calculated. This should perhaps be done
	// in some other way, because it'd be sufficient to redraw
	// only the thumb.
	layoutVScrollbar(scrollbar);
	checkThumbOverlap();
    }

    public void tagsShown(TagEvent e) {
    }

    public void tagsHidden(TagEvent e) {
    }

  
    //////////////////////////////////////////////////////////////////////
  
    // JSTT: Function to calculate Y coordinates for the newTagHint. This
    // routine is functional, but not very efficient since it uses code from
    // the layout manager and thus could and should be replaced with better
    // integrated code. Also, this function is called from this class, when
    // perhaps it should be called from TagArea. This was implemented only to
    // demonstrate the new, better way of adding tags.

    public int getThumbDrawnY() {
	Dimension sbSize = scrollbar.getSize();
	Insets sbInsets = scrollbar.getInsets();

	int tagAreaW = 16;
	
                /*
                 * Width and left edge of the buttons and thumb.
                 */
	int itemW = sbSize.width - tagAreaW - (sbInsets.left + sbInsets.right);
	int itemX = sbInsets.left;
        
                /* Nominal locations of the buttons, assuming their preferred
                 * size will fit.
                 */
	int decrButtonH = decrButton.getPreferredSize().height;
	int decrButtonY = sbInsets.top;
        
	int lesserIncrButtonH = incrButton.getPreferredSize().height;
	int lesserIncrButtonY = decrButtonY + decrButtonH;

	int incrButtonH = incrButton.getPreferredSize().height;
	int incrButtonY = sbSize.height - (sbInsets.bottom + incrButtonH);

	int lesserDecrButtonH = lesserDecrButton.getPreferredSize().height;
	int lesserDecrButtonY = incrButtonY - lesserDecrButtonH;
	
	int newTagButtonH = decrButtonH + lesserIncrButtonH;
	int newTagButtonY = sbInsets.top;
		
	int showHideButtonH = incrButtonH + lesserDecrButtonH;
	int showHideButtonY = sbSize.height -
	    (sbInsets.bottom +
	     incrButtonH + lesserDecrButtonH);
		
	/* The thumb must fit within the height left over after we
                 * subtract the preferredSize of the buttons and the insets.
                 */
	int sbInsetsH = sbInsets.top + sbInsets.bottom;
	int sbButtonsH = decrButtonH + lesserIncrButtonH +
	    lesserDecrButtonH + incrButtonH;
	float trackH = sbSize.height - (sbInsetsH + sbButtonsH);
        
                /* Compute the height and origin of the thumb.   The case
                 * where the thumb is at the bottom edge is handled specially 
                 * to avoid numerical problems in computing thumbY.  Enforce
                 * the thumbs min/max dimensions.  If the thumb doesn't
                 * fit in the track (trackH) we'll hide it later.
                 */
	float min = scrollbar.getMinimum();
	float extent = scrollbar.getVisibleAmount();
	float range = scrollbar.getMaximum() - min;
	float value = scrollbar.getValue();

	int thumbH = (range <= 0) 
	    ? getMaximumThumbSize().height : (int)(trackH * (extent / range));
	thumbH = Math.max(thumbH, getMinimumThumbSize().height);
	thumbH = Math.min(thumbH, getMaximumThumbSize().height);
        
	int thumbY = lesserDecrButtonY - thumbH;  
	if (scrollbar.getValue() < (scrollbar.getMaximum() - scrollbar.getVisibleAmount())) {
	    float thumbRange = trackH - thumbH;
	    thumbY = (int)(0.5f + (thumbRange * ((value - min) / (range - extent))));
	    thumbY +=  lesserIncrButtonY + lesserIncrButtonH;
	}

	float thumbRange = trackH - thumbH;

	return (int)(0.5f + (thumbRange * ((scrollbar.getValue() - min) / (range - extent)))) + (thumbH / 2);
    }

    /**
     * Convert model coordinates into coordinates in the scrollbar track.
     */
    public int getVisibleY(int value) {
	float range = scrollbar.getMaximum() - scrollbar.getMinimum();
	Rectangle trackBounds = getTrackBounds();
	float trackH = trackBounds.height;
	return (trackBounds.y + (int) ((trackH / range) * (float) value));
    }

    /** Check if thumb overlaps a tag in the tag area.
      * Disable newTagButton if so.
      */
    void checkThumbOverlap() {
	int offset = trackRect.y;
	Rectangle thumbRect = getThumbBounds();
	int start = thumbRect.y-offset;
	int end   = start + thumbRect.height;
	if(tagArea.areaIntersectsTag(start,end)) {
	    newTagButton.setEnabled(false);
            tagArea.setNewTagHint(-1, -1);
	}
	else {
	    newTagButton.setEnabled(true);
            tagArea.setNewTagHint(getThumbDrawnY(), scrollbar.getValue());
	}
    }
  
 
    /**
      * A listener to listen for model changes.
      *
      */
    protected class ModelListener implements ChangeListener {
	public void stateChanged(ChangeEvent e) {
	    layoutContainer(scrollbar);
	}
    }


    /**
      * Track mouse drags.
      */
    protected class TrackListener
	extends MouseAdapter implements MouseMotionListener
    {
	protected transient int offset;
	protected transient int currentMouseX, currentMouseY;
		
	public void mouseReleased(MouseEvent e)
	    {
		if(!scrollbar.isEnabled())
		    return;

		if(trackHighlight != NO_HIGHLIGHT) {
		    Rectangle r = getTrackBounds();
		    scrollbar.repaint(r.x, r.y, r.width, r.height);
		}

		trackHighlight = NO_HIGHLIGHT;
		isDragging = false;
		offset = 0;
		scrollTimer.stop();
		scrollbar.setValueIsAdjusting(false);
	    }

	// AS
	/**
	  * If the mouse is pressed above the "thumb" component
	  * then set the scrollbars value to the exact location pressed.
	  * Note that this is completely different from only scrolling up
	  * one page at a time.
	  * If the click occurs within the "thumb" area, do nothing.
	  */
	public void mousePressed(MouseEvent e) {
	    if(!scrollbar.isEnabled())
		return;

	    scrollbar.setValueIsAdjusting(true);

	    currentMouseX = e.getX();
	    currentMouseY = e.getY();
	
	    // Clicked in the Thumb area?
	    if(getThumbBounds().contains(currentMouseX, currentMouseY))	{
		switch (scrollbar.getOrientation()) {
		case Placeholder.VERTICAL:
		    offset = currentMouseY - getThumbBounds().y;
		    break;
		case Placeholder.HORIZONTAL:
		    offset = currentMouseX - getThumbBounds().x;
		    break;
		}
		isDragging = true;
		return;
	    }            
	    isDragging = false;
							
	    Dimension sbSize = scrollbar.getSize();
	    int direction = +1;

	    switch (scrollbar.getOrientation()) {
	    case Placeholder.VERTICAL:
		if (getThumbBounds().isEmpty()) {
		    int scrollbarCenter = sbSize.height / 2;
		    direction = (currentMouseY < scrollbarCenter) ? -1 : +1;
		} else {
		    int thumbY = getThumbBounds().y;
		    direction = (currentMouseY < thumbY) ? -1 : +1;
		}
		break;                    
	    case Placeholder.HORIZONTAL:
		if (getThumbBounds().isEmpty()) {
		    int scrollbarCenter = sbSize.width / 2;
		    direction = (currentMouseX < scrollbarCenter) ? -1 : +1;
		} else {
		    int thumbX = getThumbBounds().x;
		    direction = (currentMouseX < thumbX) ? -1 : +1;
		}
		break;
	    }

	    // GIST: Disable / enable to allow direct
	    // scroll-to-the-point with a click / KPL
	    if (true) {
		scrollTo(currentMouseY-getThumbBounds().height/2);
	    } else { // The default behaviour: page up / down
		scrollByBlock(direction);
		if(!getThumbBounds().contains(currentMouseX, currentMouseY))	{
		    scrollTimer.stop();
		    scrollListener.setDirection(direction);
		    scrollListener.setScrollByBlock(true);
		    scrollTimer.start();
		}	    
	    }
	    
            tagArea.setNewTagHint(getThumbDrawnY(), scrollbar.getValue());
	    checkThumbOverlap();
	}

	/** 
	  * Set the models value to the position of the top/left
	  * of the thumb relative to the origin of the track.
	  */
	public void mouseDragged(MouseEvent e) {
	    if ( !scrollbar.isEnabled() || !isDragging) {
		return;
	    }
	    
	    MouseEvent tagE = SwingUtilities.convertMouseEvent(scrollbar,e,tagArea);
	    if (tagArea.contains(new Point(tagE.getX(),tagE.getY()))) {
		// Problem: we cannot set thumb bounds directly,
		// because we do not know them - snapToY() handles the
		// tag stuff. This is now solved by redraw in
		// tagActivated().
		tagArea.snapToY(tagE.getY());
		return;
	    }
      
	    BoundedRangeModel model = scrollbar.getModel();
	    Rectangle thumbR = getThumbBounds();
	    float trackLength;
	    int thumbMin, thumbMax, thumbPos;

	    if (scrollbar.getOrientation() == Placeholder.VERTICAL) {
		thumbMin = lesserIncrButton.getY() +
		    lesserIncrButton.getHeight();
		thumbMax = lesserDecrButton.getY() -
		    getThumbBounds().height;
		thumbPos = Math.min(thumbMax,
				    Math.max(thumbMin, (e.getY() - offset)));
		setThumbBounds(thumbR.x, thumbPos, thumbR.width, thumbR.height);
		trackLength = getTrackBounds().height;
	    }
	    else {
		thumbMin = decrButton.getX() + decrButton.getWidth();
		thumbMax = incrButton.getX() - getThumbBounds().width;
		thumbPos = Math.min(thumbMax, Math.max(thumbMin, (e.getX() - offset)));
		setThumbBounds(thumbPos, thumbR.y, thumbR.width, thumbR.height);
		trackLength = getTrackBounds().width;
	    }

	    /* Set the scrollbars value.  If the thumb has reached the end of
	     * the scrollbar, then just set the value to its maximum.  Otherwise
	     * compute the value as accurately as possible.
	     */
	    if (thumbPos == thumbMax) {
		scrollbar.setValue(model.getMaximum() - model.getExtent());
	    }
	    else {
		float valueMax = model.getMaximum() - model.getExtent();
		float valueRange = valueMax - model.getMinimum();
		float thumbValue = thumbPos - thumbMin;
		float thumbRange = thumbMax - thumbMin;
		int value = (int)(0.5 + ((thumbValue / thumbRange) * valueRange));
		scrollbar.setValue(value + model.getMinimum());
	    }

	    checkThumbOverlap();
	}
		
	public void mouseMoved(MouseEvent e) {
	}
    }
        

    /**
      * Listener for cursor keys.
      */
    protected class ArrowButtonListener extends MouseAdapter {		
	// Because we are handling both mousePressed and Actions
	// we need to make sure we don't fire under both conditions.
	// (keyfocus on scrollbars causes action without mousePress
	boolean handledEvent;
    
	public void mousePressed(MouseEvent e)		{
	    if(!scrollbar.isEnabled())
		return;
      
	    int direction = (e.getSource() == incrButton ||
			     e.getSource() == lesserIncrButton) ? 1 : -1;
      
	    scrollByUnit(direction);
	    scrollTimer.stop();
	    scrollListener.setDirection(direction);
	    scrollListener.setScrollByBlock(false);
	    scrollTimer.start();

	    handledEvent = true;	
	}

	public void mouseReleased(MouseEvent e)		{
	    scrollTimer.stop();
	    handledEvent = false;
	    scrollbar.setValueIsAdjusting(false);
	}
    }

    protected class NewTagButtonListener implements ActionListener {
	public void actionPerformed(ActionEvent e) {
	    int currentValue = scrollbar.getValue();
	    tagArea.addTag(currentValue);
	}
    }

    protected class ShowHideButtonListener implements ActionListener {
	public void actionPerformed(ActionEvent e) {
	    System.err.println("toggling show/hide");
	    tagArea.setTagsVisible(showHideButton.isSelected());
	    if (showHideButton.isSelected() == true)
		showHideButton.setToolTipText("Hide tags");
	    else 
		showHideButton.setToolTipText("Show tags");
	}
    }
  
    /**
      * Listener for scrolling events intiated in the
      * ScrollPane.
      */
    protected class ScrollListener implements ActionListener {
	int direction = +1;
	boolean useBlockIncrement;
      
	public ScrollListener()	{
	    direction = +1;
	    useBlockIncrement = false;
	}
      
	public ScrollListener(int dir, boolean block)	{
	    direction = dir;
	    useBlockIncrement = block;
	}
      
	public void setDirection(int direction) { this.direction = direction; }
	public void setScrollByBlock(boolean block) {
	    this.useBlockIncrement = block;
	}
      
	public void actionPerformed(ActionEvent e) {
	    if(useBlockIncrement)	{
		scrollByBlock(direction);		
		// Stop scrolling if the thumb catches up with the mouse
		if(scrollbar.getOrientation() == Placeholder.VERTICAL)	{
		    if(direction > 0)	{
			if(getThumbBounds().y + getThumbBounds().height 
			   >= trackListener.currentMouseY)
			    ((Timer)e.getSource()).stop();
		    } else if(getThumbBounds().y <= trackListener.currentMouseY) {
			((Timer)e.getSource()).stop();
		    }
		} else {
		    if(direction > 0)	{
			if(getThumbBounds().x + getThumbBounds().width 
			   >= trackListener.currentMouseX)
			    ((Timer)e.getSource()).stop();
		    } else if(getThumbBounds().x <= trackListener.currentMouseX) {
			((Timer)e.getSource()).stop();
		    }
		}
	    } else {
		scrollByUnit(direction);
	    }
	
	    if(direction > 0 
	       && scrollbar.getValue()+scrollbar.getVisibleAmount() 
	       >= scrollbar.getMaximum())
		((Timer)e.getSource()).stop();
	    else if(direction < 0 
		    && scrollbar.getValue() <= scrollbar.getMinimum())
		((Timer)e.getSource()).stop();
	}
    }
  
  
    public class PropertyChangeHandler implements PropertyChangeListener 
    {
        public void propertyChange(PropertyChangeEvent e) {
	    String propertyName = e.getPropertyName();

	    if ("model".equals(propertyName)) {
	        BoundedRangeModel oldModel = (BoundedRangeModel)e.getOldValue();
	        BoundedRangeModel newModel = (BoundedRangeModel)e.getNewValue();
		oldModel.removeChangeListener(modelListener);
		newModel.addChangeListener(modelListener);
		scrollbar.repaint();
		scrollbar.revalidate();
	    } else if ("orientation".equals(propertyName)) {
                Integer orient = (Integer)e.getNewValue();
                if (incrButton instanceof BasicArrowButton) {
                    ((BasicArrowButton)incrButton).setDirection(orient.intValue() == HORIZONTAL?
                                                                EAST : SOUTH);
                }
                if (decrButton instanceof BasicArrowButton) {
                    ((BasicArrowButton)decrButton).setDirection(orient.intValue() == HORIZONTAL?
                                                                WEST : NORTH);
                }
            }
	}
    }

}
            
            
